@@ -2,10 +2,11 @@ import pdb
 import numpy as np
 import time
 
-from .util_ import procrustes, issparse
+from .util_ import procrustes, issparse, sparse_matrix, nearest_neighbors
 
 import scipy
-from scipy.sparse import csr_matrix
+from scipy.sparse import linalg as slinalg
+from scipy.sparse import csr_matrix, block_diag, vstack
 from sklearn.neighbors import NearestNeighbors
 from scipy.spatial.distance import pdist, squareform
 
@@ -18,43 +19,20 @@ from multiprocess import shared_memory
 def compute_Z_s_to_tear(y, s, Z_s, C, c, k):
     n_Z_s = Z_s.shape[0]
     # C_s_U_C_Z_s = (self.C[s,:]) | np.isin(self.c, Z_s)
-    C_s_U_C_Z_s = (C[s,:]) | np.any(C[Z_s,:], 0)
-    c = c[C_s_U_C_Z_s]
-    n_ = c.shape[0]
-
-    d_e_ = squareform(pdist(y[C_s_U_C_Z_s,:]))
-
-    k_ = min(k,d_e_.shape[0]-1)
-    neigh = NearestNeighbors(n_neighbors=k_,
-                             metric='precomputed',
-                             algorithm='brute')
-    neigh.fit(d_e_)
-    neigh_dist, _ = neigh.kneighbors()
-
-    epsilon = neigh_dist[:,[k_-1]]
-    Ug = d_e_ < (epsilon + 1e-12)
-
-    Utildeg = np.zeros((n_Z_s+1,n_))
-    for m in range(n_Z_s):
-        Utildeg[m,:] = np.any(Ug[c==Z_s[m],:], 0)
-
-    Utildeg[n_Z_s,:] = np.any(Ug[c==s,:], 0)
-
-    # |Utildeg_{mm'}|
-    n_Utildeg_Utildeg = np.matmul(Utildeg, Utildeg.T)
-    np.fill_diagonal(n_Utildeg_Utildeg, 0)
-
-    return Z_s[n_Utildeg_Utildeg[-1,:-1]>0]
-
+    C_s_U_C_Z_s = np.where(C[s,:] + C[Z_s,:].sum(axis=0))[1]
+    n_ = C_s_U_C_Z_s.shape[0]
+    k_ = min(k,n_-1)
+    _, neigh_ind_ = nearest_neighbors(y[C_s_U_C_Z_s,:], k_, 'euclidean')
+    U_ = sparse_matrix(neigh_ind_, np.ones(neigh_ind_.shape, dtype=bool))
+    Utilde_ = C[np.ix_(Z_s,C_s_U_C_Z_s)].dot(U_)
+    Utilde_ = vstack([Utilde_, C[s,C_s_U_C_Z_s].dot(U_)])
+    n_Utildeg_Utilde_ = Utilde_.dot(Utilde_.T) 
+    n_Utildeg_Utilde_.setdiag(False)
+    return Z_s[n_Utildeg_Utilde_[-1,:-1].nonzero()[1]].tolist()
 
 def sequential_init(seq, rho, y, is_visited_view, d, Utilde, n_Utilde_Utilde,
-                    C, c, intermed_param, global_opts, print_freq=1000,
-                    ret_contrib_of_views=False):   
+                    C, c, intermed_param, global_opts, print_freq=1000):   
     n = Utilde.shape[1]
-    if ret_contrib_of_views:
-        # Initialization contribution of view i
-        # as the points in cluster i
-        contrib_of_view = C.copy()
     # Traverse views from 2nd view
     for m in range(1,seq.shape[0]):
         if print_freq and np.mod(m, print_freq)==0:
@@ -73,27 +51,24 @@ def sequential_init(seq, rho, y, is_visited_view, d, Utilde, n_Utilde_Utilde,
                 # Compute T_s and v_s by aligning
                 # the embedding of the overlap Utilde_{sp}
                 # due to sth view with that of the pth view
-                Utilde_s_p = Utilde_s*Utilde[p,:]
+                Utilde_s_p = Utilde_s.multiply(Utilde[p,:]).nonzero()[1]
                 V_s_p = intermed_param.eval_({'view_index': s, 'data_mask': Utilde_s_p})
                 V_p_s = intermed_param.eval_({'view_index': p, 'data_mask': Utilde_s_p})
                 intermed_param.T[s,:,:], intermed_param.v[s,:] = procrustes(V_s_p, V_p_s)
-
+                
                 # Compute temporary global embedding of point in sth cluster
-                C_s = C[s,:]
+                C_s = C[s,:].indices
                 y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
-
                 # Find more views to align sth view with
-                Z_s = is_visited_view & (n_Utilde_Utilde[s,:]>0)
-                Z_s_all = np.where(Z_s)[0]
+                Z_s = n_Utilde_Utilde[s,:].multiply(is_visited_view)
+                Z_s_all = Z_s.nonzero()[1]
                 Z_s = compute_Z_s_to_tear(y, s, Z_s_all, C, c, global_opts['k'])
         # otherwise
         else:
             # Align sth view with all the views which have
             # an overlap with sth view in the ambient space
-            Z_s = is_visited_view & (n_Utilde_Utilde[s,:]>0)
-            Z_s = np.where(Z_s)[0]
-
-            Z_s = Z_s.tolist()
+            Z_s = n_Utilde_Utilde[s,:].multiply(is_visited_view)
+            Z_s = Z_s.nonzero()[1].tolist()
             # If for some reason Z_s is empty
             if len(Z_s)==0:
                 Z_s = [p]
@@ -103,15 +78,12 @@ def sequential_init(seq, rho, y, is_visited_view, d, Utilde, n_Utilde_Utilde,
         # kth point if kth point is in the sth view, else zero
         n_Utilde_s_Z_s = np.zeros(n, dtype=int)
         mu_s = np.zeros((n,d))
+        cov_s = csr_matrix((1,n), dtype=bool)
         for mp in Z_s:
-            Utilde_s_mp = Utilde_s & Utilde[mp,:]
-            if ret_contrib_of_views:
-                #overlaps[s,:] = overlaps[s,:] | Utilde_s_mp
-                #overlaps[mp,:] = overlaps[mp,:] | (Utilde[mp,:] & C[s,:])
-                contrib_of_view[s,:] = contrib_of_view[s,:] | (Utilde_s & C[mp,:])
-                
+            Utilde_s_mp = Utilde_s.multiply(Utilde[mp,:]).nonzero()[1]    
             n_Utilde_s_Z_s[Utilde_s_mp] += 1
-            mu_s[Utilde_s_mp,:] += intermed_param.eval_({'view_index': mp, 'data_mask': Utilde_s_mp})
+            mu_s[Utilde_s_mp,:] += intermed_param.eval_({'view_index': mp,
+                                                         'data_mask': Utilde_s_mp})
 
         # Compute T_s and v_s by aligning the embedding of the overlap
         # between sth view and the views in Z_s, with the centroid mu_s
@@ -129,91 +101,22 @@ def sequential_init(seq, rho, y, is_visited_view, d, Utilde, n_Utilde_Utilde,
         is_visited_view[s] = True
 
         # Compute global embedding of point in sth cluster
-        C_s = C[s,:]
+        C_s = C[s,:].indices
         y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
-    
-    if ret_contrib_of_views:
-        return y, is_visited_view, contrib_of_view
-    else:
-        return y, is_visited_view
-
-
-# Ngoc-Diep Ho, Paul Van Dooren, On the pseudo-inverse of the Laplacian of a bipartite graph
-def compute_Lpinv(Utilde):
-    B_ = Utilde.T.astype('int')
-    D_1 = np.sum(B_, axis=1, keepdims=True)
-    D_2 = np.sum(B_, axis=0, keepdims=True)
-    D_1_inv_sqrt = np.sqrt(1/D_1)
-    D_2_inv_sqrt = np.sqrt(1/D_2)
-    B_tilde = D_1_inv_sqrt*B_*D_2_inv_sqrt
-    U12,SS,VT = scipy.linalg.svd(B_tilde, full_matrices=False) # randomized svd
-    V = VT.T
-    mask = np.abs(SS-1)<1e-6
-    m_1 = np.sum(mask)
-    Sigma = np.expand_dims(SS[m_1:], 1)
-    Sigma_1 = 1/(1-Sigma**2)
-    Sigma_2 = Sigma*Sigma_1
-    U1 = U12[:,:m_1]
-    U2 = U12[:,m_1:]
-    V1 = V[:,:m_1]
-    V2 = V[:,m_1:]
-    
-    def target_proc(i):
-        if i == 0:
-            tmp = -0.75*np.matmul(U1,U1.T) + np.matmul(U2, ((Sigma_1-1))*(U2.T)) + np.eye(U12.shape[0])
-            tmp = tmp*D_1_inv_sqrt*(D_1_inv_sqrt.T)
-        elif i==1:
-            tmp = -0.25*np.matmul(U1,V1.T) + np.matmul(U2, Sigma_2*(V2.T))
-            tmp = tmp*D_1_inv_sqrt*D_2_inv_sqrt
-        elif i==2:
-            tmp = 0.25*np.matmul(V1,V1.T) + np.matmul(V2, Sigma_1*(V2.T))
-            tmp = tmp*(D_2_inv_sqrt.T)*D_2_inv_sqrt
-        return tmp, np.sum(tmp, axis=1, keepdims=True), np.sum(tmp, axis=0, keepdims=True)
-    
-    L_tilde = []
-    L_tilde_sum_1 = []
-    L_tilde_sum_0 = []
-    ################################################
-    # Parallel
-#     with mp.Pool(3) as p:
-#         tmp = p.map(target_proc, [0,1,2])
-#         for elem in tmp:
-#             L_tilde.append(elem[0])
-#             L_tilde_sum_1.append(elem[1])
-#             L_tilde_sum_0.append(elem[2])
-        
-    # Sequential
-    for i in range(3):
-        elem = target_proc(i)
-        L_tilde.append(elem[0])
-        L_tilde_sum_1.append(elem[1])
-        L_tilde_sum_0.append(elem[2])
-    ################################################
-    
-    # 1. Do not concatenate. Store blocks.
-    L_pinv_1 = np.concatenate((L_tilde[0], L_tilde[1]), axis=1)
-    L_pinv_2 = np.concatenate((L_tilde[1].T, L_tilde[2]), axis=1)
-    Lpinv = np.concatenate([L_pinv_1, L_pinv_2], axis=0)
-
-    n1 = Lpinv.shape[1]
-    v1 = np.concatenate((L_tilde_sum_1[0]+L_tilde_sum_1[1],
-                         L_tilde_sum_0[1].T+L_tilde_sum_1[2]), axis=0)
-    n0 = Lpinv.shape[0]
-    v0 = np.concatenate((L_tilde_sum_0[0]+L_tilde_sum_1[1].T,
-                         L_tilde_sum_0[1]+L_tilde_sum_0[2]), axis=1)
-    Lpinv = Lpinv - v1/n1 - v0.T/n0
-    return Lpinv
+    return y, is_visited_view
     
 # Ngoc-Diep Ho, Paul Van Dooren, On the pseudo-inverse of the Laplacian of a bipartite graph
 def compute_Lpinv_BT(Utilde, B):
     M, n = Utilde.shape
-    B_ = Utilde.T.astype('int')
-    D_1 = np.sum(B_, axis=1, keepdims=True)
-    D_2 = np.sum(B_, axis=0, keepdims=True)
+    B_ = Utilde.copy().transpose().astype('int')
+    D_1 = np.asarray(B_.sum(axis=1))
+    D_2 = np.asarray(B_.sum(axis=0))
     D_1_inv_sqrt = np.sqrt(1/D_1)
     D_2_inv_sqrt = np.sqrt(1/D_2)
-    B_tilde = D_1_inv_sqrt*B_*D_2_inv_sqrt
-    U12,SS,VT = scipy.linalg.svd(B_tilde, full_matrices=False) # randomized svd
+    B_tilde = B_.multiply(D_2_inv_sqrt).multiply(D_1_inv_sqrt)
+    # TODO: U12 is dense of size nxM
+    U12,SS,VT = scipy.linalg.svd(B_tilde.todense(), full_matrices=False)
+    # U12,SS,VT = slinalg.svds(B_tilde, k=M, solver='propack')
     V = VT.T
     mask = np.abs(SS-1)<1e-6
     m_1 = np.sum(mask)
@@ -247,68 +150,57 @@ def compute_Lpinv_BT(Utilde, B):
     temp = temp - np.mean(temp, axis=0, keepdims=True)
     return temp
 
-# def compute_CC(D, B, Lpinv):
-#     # 3. Low rank or sparse CC?
-#     #CC = D - np.matmul(B, np.matmul(Lpinv, B.T))
-#     CC = D - B.dot(B.dot(Lpinv).transpose()).transpose()
-#     CC = 0.5*(CC + CC.T)
-#     return CC
-
 def compute_CC(D, B, Lpinv_BT):
     CC = D - B.dot(Lpinv_BT)
     return 0.5*(CC + CC.T)
 
-def build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6, Lpinv_BT=None):
+def build_ortho_optim(d, Utilde, intermed_param):
     M,n = Utilde.shape
-    # B = np.zeros((M*d,n+M))
     B_row_inds = []
     B_col_inds = []
     B_vals = []
-    D = np.zeros((M*d,M*d))
+    D = []
     for i in range(M):
         X_ = intermed_param.eval_({'view_index': i,
-                                   'data_mask': Utilde[i,:]})
-        D[d*i:d*(i+1),d*i:d*(i+1)] = np.matmul(X_.T,X_)
-        # B[d*i:d*(i+1),n+i] = -np.sum(X_.T, axis=1)
-        # B[d*i:d*(i+1),np.where(Utilde[i,:])[0]] = X_.T
+                                   'data_mask': Utilde[i,:].indices})
+        D.append(np.matmul(X_.T,X_))
         row_inds = list(range(d*i,d*(i+1)))
-        col_inds = np.where(Utilde[i,:])[0].tolist()
+        col_inds = Utilde[i,:].indices.tolist()
         B_row_inds += (row_inds + np.repeat(row_inds, len(col_inds)).tolist())
         B_col_inds += (np.repeat([n+i], d).tolist() + np.tile(col_inds, d).tolist())
         B_vals += (np.sum(-X_.T, axis=1).tolist() + X_.T.flatten().tolist())
-        
-    B = csr_matrix((B_vals, (B_row_inds, B_col_inds)), shape=(M*d,n+M))
     
-    if Lpinv_BT is None:
-        print('Computing Pseudoinverse of a matrix of L of size', n+M, 'multiplied with B', flush=True)
-        Lpinv_BT = compute_Lpinv_BT(Utilde, B)
-        
-    # CC = compute_CC(D, B, Lpinv)
+    D = block_diag(D, format='csr')
+    B = csr_matrix((B_vals, (B_row_inds, B_col_inds)), shape=(M*d,n+M))
+
+    print('Computing Pseudoinverse of a matrix of L of size', n+M, 'multiplied with B', flush=True)
+    Lpinv_BT = compute_Lpinv_BT(Utilde, B)
+
     CC = compute_CC(D, B, Lpinv_BT)
     return CC, Lpinv_BT
     
 
-def compute_alignment_err(d, Utilde, intermed_param, tol = 1e-6, CC=None, Lpinv_BT=None):
-    if (CC is None) or (Lpinv is None) or (B is None):
-        CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6)
-    else:
-        CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6, Lpinv_BT=Lpinv_BT)
-    err = 0
+def compute_alignment_err(d, Utilde, intermed_param):
+    CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param)
     M,n = Utilde.shape
     CC_mask = np.tile(np.eye(d, dtype=bool), (M,M))
     err = np.sum(CC[CC_mask])
-    return err, [CC, Lpinv_BT]
-    
-def spectral_init(y, is_visited_view, d, Utilde,
-                  C, intermed_param, global_opts, print_freq=1000,
-                  tol = 1e-6):
-    CC, Lpinv, B = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6)
+    return err
+
+# Kunal N Chaudhury, Yuehaw Khoo, and Amit Singer, Global registration
+# of multiple point clouds using semidefinite programming
+def spectral_alignment(y, is_visited_view, d, Utilde,
+                      C, intermed_param, global_opts, 
+                      seq_of_intermed_views_in_cluster):
+    CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param)
     M,n = Utilde.shape
     print('Computing eigh(C,k=d)', flush=True)
     np.random.seed(42)
     v0 = np.random.uniform(0,1,CC.shape[0])
     # To find smallest eigenvalues, using shift-inverted algo with mode=normal and which='LM'
-    W_,V_ = scipy.sparse.linalg.eigsh(CC, k=d, v0=v0, sigma=0.0)
+    #W_,V_ = scipy.sparse.linalg.eigsh(CC, k=d, v0=v0, sigma=0.0)
+    # or just pass which='SM' without using sigma
+    W_,V_ = scipy.sparse.linalg.eigsh(CC, k=d, v0=v0, which='SM')
     print('Done.', flush=True)
     Wstar = np.sqrt(M)*V_.T
     
@@ -316,39 +208,52 @@ def spectral_init(y, is_visited_view, d, Utilde,
     for i in range(M):
         U_,S_,VT_ = scipy.linalg.svd(Wstar[:,d*i:d*(i+1)])
         temp_ = np.matmul(U_,VT_)
-        if (global_opts['init_algo_name'] != 'spectral') and (np.linalg.det(temp_) < 0):
+        if (global_opts['init_algo_name'] != 'spectral') and (np.linalg.det(temp_) < 0): # remove reflection
             VT_[-1,:] *= -1
             Tstar[:,i*d:(i+1)*d] = np.matmul(U_, VT_)
         else:
             Tstar[:,i*d:(i+1)*d] = temp_
     
-    Zstar = np.matmul(Tstar, np.matmul(B, Lpinv))
+    Zstar = Tstar.dot(Lpinv_BT.transpose())
     
-    for s in range(M):
-        T_s = Tstar[:,s*d:(s+1)*d].T
-        v_s = Zstar[:,n+s]
-        intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
-        intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s
-        C_s = C[s,:]
-        y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
-        is_visited_view[s] = 1
+    n_clusters = len(seq_of_intermed_views_in_cluster)
+    for i in range(n_clusters):
+        seq = seq_of_intermed_views_in_cluster[i]
+        s0 = seq[0]
+        T0T = Tstar[:,s0*d:(s0+1)*d]
+        v0 = Zstar[:,n+s0][np.newaxis,:]
+        v0TOT = np.matmul(v0, T0T)
+        is_visited_view[s0] = 1
+        for m in range(1, seq.shape[0]):
+            s = seq[m]
+            T_s = np.matmul(Tstar[:,s*d:(s+1)*d].T, T0T)
+            v_s = np.matmul(Zstar[:,n+s][np.newaxis,:], T0T) - v0TOT
+            #T_s = Tstar[:,s*d:(s+1)*d].T
+            #v_s = Zstar[:,n+s][np.newaxis,:]
+            intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
+            intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:], T_s) + v_s
+            C_s = C[s,:].indices
+            y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
+            is_visited_view[s] = 1
     
     return y, Zstar[:,:n].T, is_visited_view
 
 def sequential_final(y, d, Utilde, C, intermed_param, n_Utilde_Utilde, n_Utildeg_Utildeg,
-                     first_intermed_view_in_cluster,
-                     parents_of_intermed_views_in_cluster, cluster_of_intermed_view,
-                     global_opts):
+                     seq_of_intermed_views_in_cluster, parents_of_intermed_views_in_cluster,
+                     cluster_of_intermed_view, global_opts):
     M,n = Utilde.shape
     # Traverse over intermediate views in a random order
     seq = np.random.permutation(M)
+    is_first_view_in_cluster = np.zeros(M, dtype=bool)
+    for i in range(len(seq_of_intermed_views_in_cluster)):
+        is_first_view_in_cluster[seq_of_intermed_views_in_cluster[i][0]] = True
 
     # For a given seq, refine the global embedding
     for it1 in range(global_opts['refine_algo_max_internal_iter']):
         for s in seq.tolist():
             # Never refine s_0th intermediate view
-            if s in first_intermed_view_in_cluster:
-                C_s = C[s,:]
+            if is_first_view_in_cluster[s]:
+                C_s = C[s,:].indices
                 y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
                 continue
 
@@ -357,14 +262,14 @@ def sequential_final(y, d, Utilde, C, intermed_param, n_Utilde_Utilde, n_Utildeg
             # If to tear apart closed manifolds
             if global_opts['to_tear']:
                 # Find more views to align sth view with
-                Z_s = (n_Utilde_Utilde[s,:] > 0) & (n_Utildeg_Utildeg[s,:] > 0)
+                Z_s = n_Utilde_Utilde[s,:].multiply(n_Utildeg_Utildeg[s,:])
             # otherwise
             else:
                 # Align sth view with all the views which have
                 # an overlap with sth view in the ambient space
-                Z_s = n_Utilde_Utilde[s,:] > 0
+                Z_s = n_Utilde_Utilde[s,:]
 
-            Z_s = np.where(Z_s)[0].tolist()
+            Z_s = Z_s.nonzero()[1].tolist()
 
             if len(Z_s) == 0:
                 Z_s = parents_of_intermed_views_in_cluster[cluster_of_intermed_view[s]][s]
@@ -376,7 +281,7 @@ def sequential_final(y, d, Utilde, C, intermed_param, n_Utilde_Utilde, n_Utildeg
             n_Utilde_s_Z_s = np.zeros(n, dtype=int)
             mu_s = np.zeros((n,d))
             for mp in Z_s:
-                Utilde_s_mp = Utilde_s & Utilde[mp,:]
+                Utilde_s_mp = Utilde_s.multiply(Utilde[mp,:]).nonzero()[1]
                 n_Utilde_s_Z_s[Utilde_s_mp] += 1
                 mu_s[Utilde_s_mp,:] += intermed_param.eval_({'view_index': mp, 'data_mask': Utilde_s_mp})
 
@@ -394,20 +299,16 @@ def sequential_final(y, d, Utilde, C, intermed_param, n_Utilde_Utilde, n_Utildeg
             intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s
 
             # Compute global embedding of points in sth cluster
-            C_s = C[s,:]
+            C_s = C[s,:].indices
             y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
     return y
 
 def retraction_final(y, d, Utilde, C, intermed_param,
                      n_Utilde_Utilde, n_Utildeg_Utildeg,
-                     first_intermed_view_in_cluster,
                      parents_of_intermed_views_in_cluster,
                      cluster_of_intermed_view,
-                     global_opts, CC=None, Lpinv_BT=None):
-    if (CC is None) or (Lpinv_BT is None):
-        CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6)
-    else:
-        CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6, Lpinv=Lpinv)
+                     global_opts):
+    CC, Lpinv_BT = build_ortho_optim(d, Utilde, intermed_param)
     M,n = Utilde.shape
     n_proc = min(M,global_opts['n_proc'])
     barrier = mp.Barrier(n_proc)
@@ -494,8 +395,6 @@ def retraction_final(y, d, Utilde, C, intermed_param,
     shm_CC.close()
     shm_CC.unlink()
     
-    
-    #Zstar = np.matmul(Tstar, np.matmul(B, Lpinv))
     Zstar = Tstar.dot(Lpinv_BT.transpose())
     
     for s in range(M):
@@ -503,7 +402,7 @@ def retraction_final(y, d, Utilde, C, intermed_param,
         v_s = Zstar[:,n+s]
         intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
         intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s
-        C_s = C[s,:]
+        C_s = C[s,:].indices
         y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
 
-    return y, [CC, Lpinv_BT]
\ No newline at end of file
+    return y
\ No newline at end of file
